@@ -14,6 +14,7 @@ from .Spout import Spout
 from .utils import *
 
 from limacharlie import GLOBAL_OID
+from limacharlie import GLOBAL_UID
 from limacharlie import GLOBAL_API_KEY
 
 ROOT_URL = 'https://api.limacharlie.io'
@@ -26,7 +27,7 @@ HTTP_UNAUTHORIZED = 401
 class Manager( object ):
     '''General interface to a limacharlie.io Organization.'''
 
-    def __init__( self, oid, secret_api_key, inv_id = None, print_debug_fn = None, is_interactive = False, extra_params = {}, jwt = None ):
+    def __init__( self, oid, secret_api_key, inv_id = None, print_debug_fn = None, is_interactive = False, extra_params = {}, jwt = None, uid = None ):
         '''Create a session manager for interaction with limacharlie.io, much of the Python API relies on this object.
 
         Args:
@@ -37,12 +38,16 @@ class Manager( object ):
             is_interactive (bool): if True, the manager will provide a root investigation and Spout so that tasks sent to Sensors can be tracked in realtime automatically; requires an inv_id to be set.
             extra_params (dict): optional key / values passed to interactive spout.
             jwt (str): optionally specify a single JWT to use for authentication.
+            uid (str): a limacharlie.io user ID, if present authentication will be based on it instead of organization ID.
         '''
         # If no creds were provided, use the global ones.
         if oid is None:
             if GLOBAL_OID is None:
                 raise Exception( 'LimaCharlie "default" environment not set, please use "limacharlie login".' )
             oid = GLOBAL_OID
+        if uid is None:
+            if GLOBAL_UID is not None:
+                uid = GLOBAL_UID
         if secret_api_key is None:
             if GLOBAL_API_KEY is None and jwt is None:
                 raise Exception( 'LimaCharlie "default" environment not set, please use "limacharlie login".' )
@@ -58,6 +63,7 @@ class Manager( object ):
             if jwt is None:
                 raise LcApiException( 'Invalid secret API key, should be in UUID format.' )
         self._oid = oid
+        self._uid = uid
         self._secret_api_key = secret_api_key
         self._jwt = jwt
         self._debug = print_debug_fn
@@ -90,8 +96,13 @@ class Manager( object ):
         try:
             if self._secret_api_key is None:
                 raise Exception( 'No API key set' )
+            authData = { "secret" : self._secret_api_key }
+            if self._uid is not None:
+                authData[ 'uid' ] = self._uid
+            else:
+                authData[ 'oid' ] = self._oid
             request = urllib2.Request( API_TO_JWT_URL,
-                                       urllib.urlencode( { "oid" : self._oid, "secret" : self._secret_api_key } ) )
+                                       urllib.urlencode( authData ) )
             request.get_method = lambda: "POST"
             u = urllib2.urlopen( request )
             self._jwt = json.loads( u.read() )[ 'jwt' ]
@@ -100,16 +111,25 @@ class Manager( object ):
             self._jwt = None
             raise LcApiException( 'Failed to get JWT from API key: %s' % e )
 
-    def _restCall( self, url, verb, params ):
+    def _restCall( self, url, verb, params, altRoot = None, queryParams = None, rawBody = None, contentType = None ):
         try:
             headers = { "Authorization" : "bearer %s" % self._jwt }
 
-            url = '%s/%s/%s' % ( ROOT_URL, API_VERSION, url )
+            if altRoot is None:
+                url = '%s/%s/%s' % ( ROOT_URL, API_VERSION, url )
+            else:
+                url = '%s/%s' % ( altRoot, url )
+
+            if queryParams is not None:
+                url = '%s?%s' % ( url, urllib.urlencode( queryParams ) )
 
             request = urllib2.Request( url,
-                                       urllib.urlencode( params, doseq = True ),
+                                       rawBody if rawBody is not None else urllib.urlencode( params, doseq = True ),
                                        headers = headers )
             request.get_method = lambda: verb
+            request.add_header( 'User-Agent', 'lc-py-api' )
+            if contentType is not None:
+                request.add_header( 'Content-Type', contentType )
             u = urllib2.urlopen( request )
             try:
                 data = u.read()
@@ -132,15 +152,15 @@ class Manager( object ):
 
         return ret
 
-    def _apiCall( self, url, verb, params = {} ):
+    def _apiCall( self, url, verb, params = {}, altRoot = None, queryParams = None, rawBody = None, contentType = None ):
         if self._jwt is None:
             self._refreshJWT()
 
-        code, data = self._restCall( url, verb, params )
+        code, data = self._restCall( url, verb, params, altRoot = altRoot, queryParams = queryParams, rawBody = rawBody, contentType = contentType )
 
         if code == HTTP_UNAUTHORIZED:
             self._refreshJWT()
-            code, data = self._restCall( url, verb, params )
+            code, data = self._restCall( url, verb, params, altRoot = altRoot, queryParams = queryParams, rawBody = rawBody, contentType = contentType )
 
         if 200 != code:
             raise LcApiException( 'Api failure (%s): %s' % ( code, str( data ) ) )
@@ -166,18 +186,70 @@ class Manager( object ):
         self._is_interactive = True
         self._refreshSpout()
 
-    def testAuth( self ):
+    def testAuth( self, permissions = [] ):
         '''Tests authentication with limacharlie.io.
 
+        Args:
+            permissions (list): optional list of permissions validate we have.
+
         Returns:
             a boolean indicating whether authentication succeeded.
         '''
         try:
-            self._refreshJWT()
+            perms = None
+
+            # First make sure we have an API key or JWT.
+            if self._secret_api_key is not None:
+                try:
+                    self._refreshJWT()
+                except:
+                    return False
+            elif self._jwt is not None:
+                try:
+                    perms = self.whoAmI()
+                except:
+                    return False
+            else:
+                return False
+
+            # If there are no permissions to check, we're good since
+            # the previous part of this check made sure our auth was
+            # at least valid.
+            if permissions is None or 0 == len( permissions ):
+                return True
+
+            # We need to check the permissions we have.
+            if perms is None:
+                perms = self.whoAmI()
+            if 'user_perms' in perms:
+                # This is from a user token with permissions to multiple
+                # organizations.
+                effective = perms[ 'user_perms' ].get( self._oid, [] )
+            else:
+                # This is a machine token. Check the current OID is in there.
+                if self._oid in perms.get( 'orgs', [] ):
+                    effective = perms.get( 'perms', [] )
+                else:
+                    effective = []
+
+            # Now just check if we have them all.
+            for p in permissions:
+                if p not in effective:
+                    return False
             return True
         except:
             return False
 
+    def whoAmI( self ):
+        '''Query the API to see which organizations we are authenticated for.
+
+        Returns:
+            A list of organizations and permissions, or a dictionary of organizations with the related permissions.
+        '''
+
+        resp = self._apiCall( 'who', GET, {}, altRoot = ROOT_URL )
+        return resp
+
     def sensor( self, sid, inv_id = None ):
         '''Get a Sensor object for the specific Sensor ID.
 
